import datetime as dt
from urllib.parse import quote

from httpx import URL, AsyncClient
from pydantic import BaseModel, Field


class NationalGridESOCarbonIntensityIntensity(BaseModel):
    actual: int | None
    forecast: int


class NationalGridESOCarbonIntensityData(BaseModel):
    intensity: NationalGridESOCarbonIntensityIntensity
    from_: dt.datetime = Field(..., alias="from")


class NationalGridESOCarbonIntensityResponse(BaseModel):
    data: list[NationalGridESOCarbonIntensityData]


class NationalGridESOCarbonIntensityApiRepo:
    def __init__(self, base_url: URL):
        self._client = AsyncClient(base_url=base_url, http2=True)

    async def get_carbon_intensity(self) -> int:
        response = await self._client.get("/intensity")
        response.raise_for_status()

        parsed_response = NationalGridESOCarbonIntensityResponse.model_validate_json(
            response.content
        )

        if not parsed_response.data:
            raise ValueError("data is empty")

        first_item = parsed_response.data[0]

        return first_item.intensity.actual or first_item.intensity.forecast

    async def get_best_time_to_run_within_period(
        self, within: dt.timedelta
    ) -> dt.datetime:
        utcnow = dt.datetime.utcnow().replace(tzinfo=dt.timezone.utc)
        start_time_escaped = quote(
            utcnow.isoformat(timespec="minutes").replace("+00:00", "Z"), safe=":"
        )
        end_time_escaped = quote(
            (utcnow + within).isoformat(timespec="minutes").replace("+00:00", "Z"),
            safe=":",
        )

        response = await self._client.get(
            f"/intensity/{start_time_escaped}/{end_time_escaped}"
        )
        response.raise_for_status()

        parsed_response: NationalGridESOCarbonIntensityResponse = (
            NationalGridESOCarbonIntensityResponse.model_validate_json(response.content)
        )

        lowest_intensity = min(parsed_response.data, key=lambda x: x.intensity.forecast)

        return lowest_intensity.from_



import datetime as dt
from pathlib import Path
from typing import Optional

import typer
from httpx import URL
from typing_extensions import Annotated

from src.cli.command.check import run as check_run
from src.cli.command.schedule import run as schedule_run
from src.cli.parsers import http_or_https_url, human_readable_duration
from src.repos.carbon_intensity.protocol import (
    CurrentIntensityDataSource,
    ForecastLowIntensityDateDataSource,
    MissingParameterError,
    make_current_intensity_repo,
    make_forecast_low_intensity_date_repo,
)

NATIONAL_GRID_ESO_CARBON_INTENSITY_API_BASE_URL: URL = URL(
    "https://api.carbonintensity.org.uk"
)
CO2_SIGNAL_API_BASE_URL: URL = URL("https://api.co2signal.com")
FILE_FOR_INTENSITY_READING: str = ".carbon_intensity"

app = typer.Typer()


@app.command(help="Check the current carbon intensity.")
def check(
    max_carbon_intensity: Annotated[
        int,
        typer.Option(
            envvar="MAX_CARBON_INTENSITY",
            help="Set the max carbon intensity in gCO2eq/kWh.",
        ),
    ],
    advise_only: Annotated[
        bool,
        typer.Option(
            envvar="ADVISE_ONLY",
            help="Do not exit with an error if the carbon intensity is above the max carbon intensity.",
        ),
    ] = False,
    data_source: Annotated[
        CurrentIntensityDataSource,
        typer.Option(
            case_sensitive=False,
            envvar="DATA_SOURCE",
            help="Where to read carbon intensity data from",
        ),
    ] = CurrentIntensityDataSource.NATIONAL_GRID_ESO_CARBON_INTENSITY,
    from_file_carbon_intensity_file_path: Annotated[
        Path,
        typer.Option(
            envvar="FROM_FILE_CARBON_INTENSITY_FILE_PATH",
            help="File to read carbon intensity from in file mode",
        ),
    ] = Path(FILE_FOR_INTENSITY_READING),
    national_grid_eso_carbon_intensity_api_base_url: Annotated[
        URL,
        typer.Option(
            envvar="NATIONAL_GRID_ESO_CARBON_INTENSITY_API_BASE_URL",
            help="URL for the National Grid ESO Carbon Intensity API",
            parser=http_or_https_url,
        ),
    ] = NATIONAL_GRID_ESO_CARBON_INTENSITY_API_BASE_URL,
    co2_signal_carbon_intensity_api_base_url: Annotated[
        URL,
        typer.Option(
            envvar="CO2_SIGNAL_API_BASE_URL",
            help="URL for the CO2 Signal api",
            parser=http_or_https_url,
        ),
    ] = CO2_SIGNAL_API_BASE_URL,
    co2_signal_api_key: Annotated[
        Optional[str],
        typer.Option(
            envvar="CO2_SIGNAL_API_KEY",
            help="Api key for the CO2 Signal api, required in CO2 Signal mode",
        ),
    ] = None,
    co2_signal_country_code: Annotated[
        Optional[str],
        typer.Option(
            envvar="CO2_SIGNAL_COUNTRY_CODE",
            help="Country code to get the carbon intensity from CO2 Signal api",
        ),
    ] = None,
) -> None:
    try:
        intensity_repo = make_current_intensity_repo(
            co2_signal_api_key,
            co2_signal_carbon_intensity_api_base_url,
            co2_signal_country_code,
            data_source,
            from_file_carbon_intensity_file_path,
            national_grid_eso_carbon_intensity_api_base_url,
        )
    except MissingParameterError as e:
        typer.echo(f"{e}")
        raise typer.Exit(1) from e

    check_run(
        max_carbon_intensity,
        advise_only,
        intensity_repo,
    )


@app.command(help="Find the lowest carbon time.")
def schedule(
    within: Annotated[
        dt.timedelta,
        typer.Option(
            envvar="WITHIN",
            help="Time period to predict the lowest intensity within",  #
            parser=human_readable_duration,
        ),
    ],
    data_source: Annotated[
        ForecastLowIntensityDateDataSource,
        typer.Option(
            case_sensitive=False,
            envvar="DATA_SOURCE",
            help="Where to read carbon intensity data from",
        ),
    ] = ForecastLowIntensityDateDataSource.NATIONAL_GRID_ESO_CARBON_INTENSITY,
    from_file_carbon_intensity_file_path: Annotated[
        Path,
        typer.Option(
            envvar="FROM_FILE_CARBON_INTENSITY_FILE_PATH",
            help="File to read carbon intensity from in file mode",
        ),
    ] = Path(FILE_FOR_INTENSITY_READING),
    national_grid_eso_carbon_intensity_api_base_url: Annotated[
        URL,
        typer.Option(
            envvar="NATIONAL_GRID_ESO_CARBON_INTENSITY_API_BASE_URL",
            help="URL for the National Grid ESO Carbon Intensity API",
            parser=http_or_https_url,
        ),
    ] = NATIONAL_GRID_ESO_CARBON_INTENSITY_API_BASE_URL,
) -> None:
    intensity_repo = make_forecast_low_intensity_date_repo(
        data_source,
        from_file_carbon_intensity_file_path,
        national_grid_eso_carbon_intensity_api_base_url,
    )
    schedule_run(
        within,
        intensity_repo,
    )


def run() -> None:
    app()


if __name__ == "__main__":
    run()



import datetime as dt


class InMemoryCarbonIntensityRepo:
    def __init__(self, carbon_intensity: list[tuple[dt.datetime, int]]) -> None:
        self._carbon_intensity = carbon_intensity

    async def get_carbon_intensity(self) -> int:
        intensity = self._carbon_intensity
        utcnow = dt.datetime.utcnow().replace(tzinfo=dt.timezone.utc)

        intensity_in_present = filter(lambda x: x[0] <= utcnow, intensity)
        max_date = max(intensity_in_present, key=lambda x: x[0])

        return max_date[1]

    async def get_best_time_to_run_within_period(
        self, within: dt.timedelta
    ) -> dt.datetime:
        intensity = self._carbon_intensity
        utcnow = dt.datetime.utcnow().replace(tzinfo=dt.timezone.utc)

        intensity_in_future = filter(lambda x: x[0] > utcnow, intensity)
        intensity_within_duratation = filter(
            lambda x: (x[0] - utcnow) <= within, intensity_in_future
        )
        cleanest_time = min(intensity_within_duratation, key=lambda x: x[1])

        return cleanest_time[0]



import datetime as dt
from typing import Self

from dateparser import parse
from httpx import URL


class ArgumentParserError(ValueError):
    @classmethod
    def from_invalid_duration(cls, given_duration: str) -> Self:
        return cls(f"Could not parse {given_duration} as a human readable duration")

    @classmethod
    def from_invalid_url(cls, url: str) -> Self:
        return cls(f"{url} is not a valid http or https URL")


def http_or_https_url(url: str) -> URL:
    u = URL(url)

    if u.scheme not in ("http", "https"):
        raise ArgumentParserError.from_invalid_url(url)

    return u


def human_readable_duration(given_duration: str) -> dt.timedelta:
    anchor_time = dt.datetime.now()

    time = parse(given_duration, settings={"RELATIVE_BASE": anchor_time})

    if time is None:
        raise ArgumentParserError.from_invalid_duration(given_duration)

    return anchor_time - time



import datetime as dt

import typer

from src.cli.async_helper import async_to_sync
from src.repos.carbon_intensity.protocol import ForecastLowIntensityDateRepo


@async_to_sync
async def run(
    within: dt.timedelta, intensity_repo: ForecastLowIntensityDateRepo
) -> None:
    best_time = await intensity_repo.get_best_time_to_run_within_period(within)
    typer.echo(best_time.isoformat())
    return



import datetime as dt
from enum import StrEnum
from pathlib import Path
from typing import Protocol, Self

from httpx import URL

from src.repos.carbon_intensity.co2_signal import CO2SignalCarbonIntensityRepo
from src.repos.carbon_intensity.file import FromFileCarbonIntensityRepo
from src.repos.carbon_intensity.national_grid_eso import (
    NationalGridESOCarbonIntensityApiRepo,
)


class CurrentCarbonIntensityRepo(Protocol):
    async def get_carbon_intensity(self) -> int:
        ...


class CurrentIntensityDataSource(StrEnum):
    FILE = "file"
    NATIONAL_GRID_ESO_CARBON_INTENSITY = "national-grid-eso-carbon-intensity"
    CO2_SIGNAL = "co2-signal"


class ForecastLowIntensityDateRepo(Protocol):
    async def get_best_time_to_run_within_period(
        self, within: dt.timedelta
    ) -> dt.datetime:
        ...


class ForecastLowIntensityDateDataSource(StrEnum):
    FILE = "file"
    NATIONAL_GRID_ESO_CARBON_INTENSITY = "national-grid-eso-carbon-intensity"


class MissingParameterError(ValueError):
    @classmethod
    def from_missing_country_code(cls) -> Self:
        return cls("No country code provided to CO2 Signal Api.")

    @classmethod
    def from_missing_api_key(cls) -> Self:
        return cls("No API key found for CO2 Signal API.")


def make_current_intensity_repo(
    co2_signal_api_key: str | None,
    co2_signal_carbon_intensity_api_base_url: URL,
    co2_signal_country_code: str | None,
    data_source: CurrentIntensityDataSource,
    from_file_carbon_intensity_file_path: Path,
    national_grid_eso_carbon_intensity_api_base_url: URL,
) -> CurrentCarbonIntensityRepo:
    intensity_repo: CurrentCarbonIntensityRepo = FromFileCarbonIntensityRepo(
        from_file_carbon_intensity_file_path
    )
    match data_source:
        case CurrentIntensityDataSource.FILE:
            intensity_repo = intensity_repo
        case CurrentIntensityDataSource.NATIONAL_GRID_ESO_CARBON_INTENSITY:
            intensity_repo = NationalGridESOCarbonIntensityApiRepo(
                base_url=national_grid_eso_carbon_intensity_api_base_url
            )
        case CurrentIntensityDataSource.CO2_SIGNAL:
            if not co2_signal_country_code:
                raise MissingParameterError.from_missing_country_code()

            if not co2_signal_api_key:
                raise MissingParameterError.from_missing_api_key()

            intensity_repo = CO2SignalCarbonIntensityRepo(
                base_url=co2_signal_carbon_intensity_api_base_url,
                api_key=co2_signal_api_key,
                country_code=co2_signal_country_code,
            )
    return intensity_repo


def make_forecast_low_intensity_date_repo(
    data_source: ForecastLowIntensityDateDataSource,
    from_file_carbon_intensity_file_path: Path,
    national_grid_eso_carbon_intensity_api_base_url: URL,
) -> ForecastLowIntensityDateRepo:
    intensity_repo: ForecastLowIntensityDateRepo = FromFileCarbonIntensityRepo(
        from_file_carbon_intensity_file_path
    )
    match data_source:
        case ForecastLowIntensityDateDataSource.FILE:
            intensity_repo = intensity_repo
        case ForecastLowIntensityDateDataSource.NATIONAL_GRID_ESO_CARBON_INTENSITY:
            intensity_repo = NationalGridESOCarbonIntensityApiRepo(
                base_url=national_grid_eso_carbon_intensity_api_base_url
            )
    return intensity_repo



import typer

from src.cli.async_helper import async_to_sync
from src.repos.carbon_intensity.protocol import CurrentCarbonIntensityRepo

SUCCESS_TEMPLATE: str = "Carbon intensity is {carbon_intensity} gCO2eq/kWh, which is below or equal to the max of {max_carbon_intensity} gCO2eq/kWh"
FAILURE_TEMPLATE: str = "Carbon intensity is {carbon_intensity} gCO2eq/kWh, which is above the max of {max_carbon_intensity} gCO2eq/kWh"


@async_to_sync
async def run(
    max_carbon_intensity: int,
    advise_only: bool,
    intensity_repo: CurrentCarbonIntensityRepo,
) -> None:
    carbon_intensity = await intensity_repo.get_carbon_intensity()
    if carbon_intensity > max_carbon_intensity:
        typer.echo(
            FAILURE_TEMPLATE.format(
                max_carbon_intensity=max_carbon_intensity,
                carbon_intensity=carbon_intensity,
            )
        )
        raise typer.Exit(1 if not advise_only else 0)
    typer.echo(
        SUCCESS_TEMPLATE.format(
            max_carbon_intensity=max_carbon_intensity, carbon_intensity=carbon_intensity
        )
    )



import csv
import datetime as dt
from pathlib import Path


class FromFileCarbonIntensityRepo(object):
    def __init__(self, file_path: Path) -> None:
        self._file_path = file_path

    async def get_carbon_intensity(self) -> int:
        parsed_rows = self._read_csv()
        utcnow = dt.datetime.utcnow().replace(tzinfo=dt.timezone.utc)

        intensity_in_present = filter(lambda x: x[0] <= utcnow, parsed_rows)
        max_date = max(intensity_in_present, key=lambda x: x[0])

        return max_date[1]

    def _read_csv(self) -> list[tuple[dt.datetime, int]]:
        with self._file_path.open("r") as file:
            raw_rows = csv.reader(file)

            return [
                (
                    dt.datetime.fromisoformat(row[0]).replace(tzinfo=dt.timezone.utc),
                    int(row[1]),
                )
                for row in raw_rows
            ]

    async def get_best_time_to_run_within_period(
        self, within: dt.timedelta
    ) -> dt.datetime:
        parsed_rows = self._read_csv()
        utcnow = dt.datetime.utcnow().replace(tzinfo=dt.timezone.utc)

        intensity_in_future = filter(lambda x: x[0] > utcnow, parsed_rows)
        intensity_within_duratation = filter(
            lambda x: (x[0] - utcnow) <= within, intensity_in_future
        )
        cleanest_time = min(intensity_within_duratation, key=lambda x: x[1])

        return cleanest_time[0]



import asyncio
from typing import Callable, Coroutine, ParamSpec, TypeVar

P = ParamSpec("P")
R = TypeVar("R")


def async_to_sync(f: Callable[P, Coroutine[None, None, R]]) -> Callable[P, R]:
    def inner(*args: P.args, **kwargs: P.kwargs) -> R:
        return asyncio.run(f(*args, **kwargs))

    return inner



from typing import Generator

import httpx
from httpx import URL, AsyncClient, Request, Response
from pydantic import BaseModel


class CO2SignalCarbonIntensityData(BaseModel):
    carbonIntensity: int


class CO2SignalCarbonIntensityResponse(BaseModel):
    data: CO2SignalCarbonIntensityData


class CO2SignalCarbonIntensityRepo:
    def __init__(self, base_url: URL, api_key: str, country_code: str):
        self._client = AsyncClient(
            base_url=base_url, http2=True, auth=CO2SignalAuthClient(api_key=api_key)
        )
        self._country_code = country_code

    async def get_carbon_intensity(self) -> int:
        response = await self._client.get(
            "/v1/latest", params={"countryCode": self._country_code}
        )
        response.raise_for_status()

        parsed_response = CO2SignalCarbonIntensityResponse.model_validate_json(
            response.content
        )

        return parsed_response.data.carbonIntensity


class CO2SignalAuthClient(httpx.Auth):
    def __init__(self, api_key: str):
        self._api_key = api_key

    def auth_flow(self, request: Request) -> Generator[Request, Response, None]:
        request.headers["auth-token"] = self._api_key
        yield request
